@@ -3,14 +3,14 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2003 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2009 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  * 
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  * 
  * GAWK is distributed in the hope that it will be useful,
@@ -20,7 +20,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 /* ------------------------------ Includes ------------------------------ */
@@ -41,6 +41,10 @@
 #define _GNU_SOURCE	1	/* enable GNU extensions */
 #endif /* _GNU_SOURCE */
 
+#if defined(_TANDEM_SOURCE) && ! defined(_SCO_DS)
+#define _XOPEN_SOURCE_EXTENDED 1
+#endif
+
 #include <stdio.h>
 #include <assert.h>
 #ifdef HAVE_LIMITS_H
@@ -60,6 +64,7 @@
 #endif
 
 #if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
+#define CAN_USE_STDARG_H 1
 #include <stdarg.h>
 #else
 #include <varargs.h>
@@ -73,13 +78,38 @@ extern int errno;
 #ifdef HAVE_SIGNUM_H
 #include <signum.h>
 #endif
-#if defined(HAVE_MBRLEN) && defined(HAVE_MBRTOWC) && defined(HAVE_WCHAR_H) && defined(HAVE_WCTYPE_H)
+
+#ifndef NO_MBSUPPORT
+#include "mbsupport.h" /* defines MBS_SUPPORT */
+#endif
+
+#if defined(MBS_SUPPORT)
 /* We can handle multibyte strings.  */
-#define MBS_SUPPORT
 #include <wchar.h>
 #include <wctype.h>
 #endif
 
+#ifdef STDC_HEADERS
+#include <float.h>
+#endif
+
+#undef CHARBITS
+#undef INTBITS
+
+#if !defined(ZOS_USS)
+#if HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
+#if HAVE_STDINT_H
+# include <stdint.h>
+#endif
+#endif /* !ZOS_USS */
+ 
+#if defined(_MSC_VER)
+/* for read()/close() in use replace.c */
+#include <io.h>
+#endif
+
 /* ----------------- System dependencies (with more includes) -----------*/
 
 /* This section is the messiest one in the file, not a lot that can be done */
@@ -185,6 +215,14 @@ extern int errno;
 #include <unistd.h>
 #endif	/* HAVE_UNISTD_H */
 
+#ifdef VMS
+#include "vms/redirect.h"
+#endif  /*VMS*/
+
+#ifdef atarist
+#include "unsupported/atari/redirect.h"
+#endif
+
 #ifndef HAVE_VPRINTF
 /* if you don't have vprintf, try this and cross your fingers. */
 #ifdef	HAVE_DOPRNT
@@ -195,6 +233,11 @@ lose
 #endif	/* not HAVE_DOPRNT */
 #endif	/* HAVE_VPRINTF */
 
+#ifndef HAVE_SNPRINTF
+/* will use replacement version */
+extern int snprintf P((char *restrict buf, size_t len, const char *restrict fmt, ...));
+#endif
+
 #ifndef HAVE_SETLOCALE
 #define setlocale(locale, val)	/* nothing */
 #endif /* HAVE_SETLOCALE */
@@ -203,21 +246,16 @@ lose
    this is a hack but it gives us the right semantics */
 #define lintwarn (*(set_loc(__FILE__, __LINE__),lintfunc))
 
-#ifdef VMS
-#include "vms/redirect.h"
-#endif  /*VMS*/
-
-#ifdef atarist
-#include "unsupported/atari/redirect.h"
-#endif
-
-#define RE_TRANSLATE_TYPE const char *
 #define	GNU_REGEX
 #ifdef GNU_REGEX
 #include "regex.h"
+#include "dfa.h"
 typedef struct Regexp {
 	struct re_pattern_buffer pat;
 	struct re_registers regs;
+	struct dfa dfareg;
+	short dfa;
+	short has_anchor;	/* speed up of avoid_dfa kludge, temporary */
 } Regexp;
 #define	RESTART(rp,s)	(rp)->regs.start[0]
 #define	REEND(rp,s)	(rp)->regs.end[0]
@@ -225,9 +263,12 @@ typedef struct Regexp {
 #define	SUBPATEND(rp,s,n)	(rp)->regs.end[n]
 #define	NUMSUBPATS(rp,s)	(rp)->regs.num_regs
 #endif	/* GNU_REGEX */
+/* regexp matching flags: */
+#define RE_NEED_START	1	/* need to know start/end of match */
+#define RE_NO_BOL	2	/* not allowed to match ^ in regexp */
 
 /* Stuff for losing systems. */
-#ifdef STRTOD_NOT_C89
+#if !defined(HAVE_STRTOD)
 extern double gawk_strtod();
 #define strtod gawk_strtod
 #endif
@@ -256,11 +297,28 @@ extern double gawk_strtod();
 #define __extension__
 #endif
 
+/* this is defined by Windows32 extension libraries. It must be added to
+ * every variable which is exported (including function pointers) */
+#if defined(WIN32_EXTENSION) && !defined(ATTRIBUTE_EXPORTED)
+# define ATTRIBUTE_EXPORTED __declspec(dllimport)
+#else
+# define ATTRIBUTE_EXPORTED
+#endif
+
+
 /* ------------------ Constants, Structures, Typedefs  ------------------ */
 
+#if 0
+/* This isn't right. Figure it out the right way for the next release */
 #ifndef AWKNUM
-#define AWKNUM	double
+#ifdef LDBL_MANT_DIG
+#define AWKNUM	long double
+#else
 #endif
+#endif
+#endif
+
+#define AWKNUM	double
 
 #ifndef TRUE
 /* a bit hackneyed, but what the heck */
@@ -317,6 +375,7 @@ typedef enum nodevals {
 	Node_assign_plus,
 	Node_assign_minus,
 	Node_assign_exp,
+	Node_assign_concat,
 
 	/* boolean binaries   lnode and rnode are expressions */
 	Node_and,
@@ -339,12 +398,17 @@ typedef enum nodevals {
 	Node_rule_list,		/* lnode is a rule, rnode is rest of list */
 	Node_rule_node,		/* lnode is pattern, rnode is statement */
 	Node_statement_list,	/* lnode is statement, rnode is more list */
+	Node_switch_body,	/* lnode is the case list, rnode is default list */
+	Node_case_list,	/* lnode is the case, rnode is a statement list */
 	Node_if_branches,	/* lnode is to run on true, rnode on false */
 	Node_expression_list,	/* lnode is an exp, rnode is more list */
 	Node_param_list,	/* lnode is a variable, rnode is more list */
 
 	/* keywords */
 	Node_K_if,		/* lnode is conditonal, rnode is if_branches */
+	Node_K_switch,		/* lnode is switch value, rnode is body of case statements */
+	Node_K_case,		/* lnode is case value, rnode is stuff to run */
+	Node_K_default,		/* lnode is empty, rnode is stuff to run */
 	Node_K_while,		/* lnode is condtional, rnode is stuff to run */
 	Node_K_for,		/* lnode is for_struct, rnode is stuff to run */
 	Node_K_arrayfor,	/* lnode is for_struct, rnode is stuff to run */
@@ -372,8 +436,9 @@ typedef enum nodevals {
 	Node_redirect_twoway,	/* subnode is where to redirect */
 
 	/* Variables */
-	Node_var,		/* rnode is value, lnode is array stuff */
-	Node_var_array,		/* array is ptr to elements, asize num of eles */
+	Node_var_new,		/* newly created variable, may become an array */
+	Node_var,		/* scalar variable, lnode is value */
+	Node_var_array,		/* array is ptr to elements, table_size num of eles */
 	Node_val,		/* node is a value - type in flags */
 
 	/* Builtins   subnode is explist to work on, builtin is func to call */
@@ -416,6 +481,7 @@ typedef enum nodevals {
 	Node_ORS,
 	Node_RS,
 	Node_TEXTDOMAIN,
+	Node_SUBSEP,
 	Node_final		/* sentry value, not legal */
 } NODETYPE;
 
@@ -447,9 +513,9 @@ typedef struct exp_node {
 			char *name;
 			short number;
 			unsigned long reflags;
-#				define	CASE	1
-#				define	CONST	2
-#				define	FS_DFLT	4
+#				define	CASE		1
+#				define	CONSTANT	2
+#				define	FS_DFLT		4
 		} nodep;
 		struct {
 			AWKNUM fltnum;	/* this is here for optimal packing of
@@ -459,6 +525,10 @@ typedef struct exp_node {
 			size_t slen;
 			long sref;
 			int idx;
+#ifdef MBS_SUPPORT
+			wchar_t *wsp;
+			size_t wslen;
+#endif
 		} val;
 		struct {
 			struct exp_node *next;
@@ -466,6 +536,7 @@ typedef struct exp_node {
 			size_t length;
 			struct exp_node *value;
 			long ref;
+			size_t code;
 		} hash;
 #define	hnext	sub.hash.next
 #define	hname	sub.hash.name
@@ -477,6 +548,7 @@ typedef struct exp_node {
 #define ahname_len	sub.hash.length
 #define	ahvalue	sub.hash.value
 #define ahname_ref	sub.hash.ref
+#define	ahcode	sub.hash.code
 	} sub;
 	NODETYPE type;
 	unsigned short flags;
@@ -490,12 +562,13 @@ typedef struct exp_node {
 #		define	MAYBE_NUM 128	/* user input: if NUMERIC then
 					 * a NUMBER */
 #		define	ARRAYMAXED 256	/* array is at max size */
-#		define	SCALAR     512	/* used as scalar, can't be array */
-#		define	FUNC	1024	/* this parameter is really a
-					 * function name; see awk.y */
-#		define	FIELD	2048	/* this is a field */
-#		define	INTLSTR	4096	/* use localized version */
-#		define	UNINITIALIZED	8192	/* value used before set */
+#		define	FUNC	512	/* this parameter is really a
+					 * function name; see awkgram.y */
+#		define	FIELD	1024	/* this is a field */
+#		define	INTLSTR	2048	/* use localized version */
+#ifdef MBS_SUPPORT
+#		define	WSTRCUR	4096	/* wide str value is current */
+#endif
 } NODE;
 
 #define vname sub.nodep.name
@@ -519,6 +592,7 @@ typedef struct exp_node {
 #define re_flags sub.nodep.reflags
 #define re_text lnode
 #define re_exp	sub.nodep.x.extra
+#define	re_cnt	sub.nodep.number
 
 #define forloop	rnode->sub.nodep.r.hd
 
@@ -527,14 +601,22 @@ typedef struct exp_node {
 #define stref	sub.val.sref
 #define	stfmt	sub.val.idx
 
+#define wstptr	sub.val.wsp
+#define wstlen	sub.val.wslen
+
 #define numbr	sub.val.fltnum
 
+/* Node_var: */
 #define var_value lnode
+
+/* Node_var_array: */
 #define var_array sub.nodep.r.av
 #define array_size sub.nodep.l.ll
 #define table_size sub.nodep.x.xl
 
+/* Node_array_ref: */
 #define orig_array sub.nodep.x.extra
+#define prev_array rnode
 
 #define printf_count sub.nodep.x.xl
 
@@ -560,9 +642,13 @@ typedef struct iobuf {
 	size_t readsize;        /* set from fstat call */
 	size_t size;            /* buffer size */
 	ssize_t count;          /* amount read last time */
-	size_t total;           /* total num chars read */
 	size_t scanoff;         /* where we were in the buffer when we had
 				   to regrow/refill */
+
+	void *opaque;		/* private data for open hooks */
+	int (*get_record) P((char **out, struct iobuf *, int *errcode));
+	void (*close_func) P((struct iobuf *));		/* open and close hooks */
+
 	int flag;
 #		define	IOP_IS_TTY	1
 #		define	IOP_IS_INTERNAL	2
@@ -570,6 +656,7 @@ typedef struct iobuf {
 #		define	IOP_NOFREE_OBJ	8
 #               define  IOP_AT_EOF      16
 #               define  IOP_CLOSED      32
+#               define  IOP_AT_START    64
 } IOBUF;
 
 typedef void (*Func_ptr) P((void));
@@ -633,7 +720,7 @@ struct flagtab {
 #ifndef LONG_MIN
 #define LONG_MIN ((long)(-LONG_MAX - 1L))
 #endif
-#define HUGE    LONG_MAX 
+#define UNLIMITED    LONG_MAX 
 
 /* -------------------------- External variables -------------------------- */
 /* gawk builtin variables */
@@ -649,7 +736,7 @@ extern char *ORS;
 extern int ORSlen;
 extern char *OFMT;
 extern char *CONVFMT;
-extern int CONVFMTidx;
+ATTRIBUTE_EXPORTED extern int CONVFMTidx;
 extern int OFMTidx;
 extern char *TEXTDOMAIN;
 extern NODE *BINMODE_node, *CONVFMT_node, *FIELDWIDTHS_node, *FILENAME_node;
@@ -657,15 +744,21 @@ extern NODE *FNR_node, *FS_node, *IGNORECASE_node, *NF_node;
 extern NODE *NR_node, *OFMT_node, *OFS_node, *ORS_node, *RLENGTH_node;
 extern NODE *RSTART_node, *RS_node, *RT_node, *SUBSEP_node, *PROCINFO_node;
 extern NODE *LINT_node, *ERRNO_node, *TEXTDOMAIN_node;
-extern NODE **stack_ptr;
+ATTRIBUTE_EXPORTED extern NODE **stack_ptr;
 extern NODE *Nnull_string;
+extern NODE *Null_field;
 extern NODE **fields_arr;
 extern int sourceline;
 extern char *source;
 extern NODE *expression_value;
 
 #if __GNUC__ < 2
-extern NODE *_t;	/* used as temporary in tree_eval */
+# if defined(WIN32_EXTENSION)
+static
+# else
+extern
+#endif
+NODE *_t;	/* used as temporary in tree_eval */
 #endif
 
 extern NODE *nextfree;
@@ -677,6 +770,8 @@ extern int do_intl;
 extern int do_non_decimal_data;
 extern int do_dump_vars;
 extern int do_tidy_mem;
+extern int do_optimize;
+extern int use_lc_numeric;
 extern int in_begin_rule;
 extern int in_end_rule;
 extern int whiny_users;
@@ -684,11 +779,13 @@ extern int whiny_users;
 #define do_lint 0
 #define do_lint_old 0
 #else
-extern int do_lint;
+ATTRIBUTE_EXPORTED extern int do_lint;
 extern int do_lint_old;
 #endif
 #ifdef MBS_SUPPORT
 extern int gawk_mb_cur_max;
+#else
+extern const int gawk_mb_cur_max;
 #endif
 
 #if defined (HAVE_GETGROUPS) && defined(NGROUPS_MAX) && NGROUPS_MAX > 0
@@ -696,31 +793,35 @@ extern GETGROUPS_T *groupset;
 extern int ngroups;
 #endif
 
+#ifdef HAVE_LOCALE_H
+extern struct lconv loc;
+#endif /* HAVE_LOCALE_H */
+
 extern const char *myname;
 
 extern char quote;
 extern char *defpath;
 extern char envsep;
 
-extern const char casetable[];	/* for case-independent regexp matching */
+extern char casetable[];	/* for case-independent regexp matching */
 
 /* ------------------------- Pseudo-functions ------------------------- */
 
 #define is_identchar(c)		(isalnum(c) || (c) == '_')
-#define isnondecimal(str)	(((str)[0]) == '0' && (ISDIGIT((str)[1]) \
-					|| (str)[1] == 'x' || (str)[1] == 'X'))
+
+#define var_uninitialized(n)	((n)->var_value == Nnull_string)
 
 #ifdef MPROF
-#define	getnode(n)	emalloc((n), NODE *, sizeof(NODE), "getnode"), (n)->flags = UNINITIALIZED, (n)-exec_count = 0;
+#define	getnode(n)	emalloc((n), NODE *, sizeof(NODE), "getnode"), (n)->flags = 0, (n)->exec_count = 0;
 #define	freenode(n)	free(n)
 #else	/* not MPROF */
 #define	getnode(n)	if (nextfree) n = nextfree, nextfree = nextfree->nextp;\
 			else n = more_nodes()
 #ifndef NO_PROFILING
-#define	freenode(n)	((n)->flags = UNINITIALIZED,\
+#define	freenode(n)	((n)->flags = 0,\
 		(n)->exec_count = 0, (n)->nextp = nextfree, nextfree = (n))
 #else /* not PROFILING */
-#define	freenode(n)	((n)->flags = UNINITIALIZED,\
+#define	freenode(n)	((n)->flags = 0,\
 		(n)->nextp = nextfree, nextfree = (n))
 #endif	/* not PROFILING */
 #endif	/* not MPROF */
@@ -746,28 +847,31 @@ extern const char casetable[];	/* for case-independent regexp matching */
 #define dupnode(n)	r_dupnode(n)
 #endif	/* GAWKDEBUG */
 
+#define get_array(t)	get_actual(t, TRUE)	/* allowed to die fatally */
+#define get_param(t)	get_actual(t, FALSE)	/* not allowed */
+
 #ifdef MEMDEBUG
 #undef freenode
 #define	get_lhs(p, a, r)	r_get_lhs((p), (a), (r))
 #define	m_tree_eval(t, iscond)	r_tree_eval(t, iscond)
 #else
 #define	get_lhs(p, a, r) ((p)->type == Node_var && \
-			 ((p)->flags & UNINITIALIZED) == 0 && (r) ? \
-			  (&(p)->var_value): \
+			  ! var_uninitialized(p) ? \
+			  (&(p)->var_value) : \
 			 r_get_lhs((p), (a), (r)))
+#define TREE_EVAL_MACRO 1
 #if __GNUC__ >= 2
 #define	m_tree_eval(t, iscond) __extension__ \
                         ({NODE * _t = (t);                 \
-			   if (_t == NULL)                 \
-			       _t = Nnull_string;          \
-			   else {                          \
+			       if (_t == (NODE*)NULL)      \
+					cant_happen();     \
 			       switch(_t->type) {          \
 			       case Node_val:              \
 				   if (_t->flags&INTLSTR)  \
 					_t = r_force_string(_t); \
 				   break;                  \
 			       case Node_var:              \
-				   if ((_t->flags & UNINITIALIZED) == 0) { \
+				   if (! var_uninitialized(_t)) { \
 				       _t = _t->var_value;     		   \
 				       break;                  		   \
 				   }					   \
@@ -776,17 +880,16 @@ extern const char casetable[];	/* for case-independent regexp matching */
 				   _t = r_tree_eval(_t, iscond);\
 				   break;                  \
 			       }                           \
-			   }                               \
 			   _t;})
 #else
-#define	m_tree_eval(t, iscond)	(_t = (t), _t == NULL ? Nnull_string : \
+#define	m_tree_eval(t, iscond)	(_t = (t), _t == (NODE*)NULL ? (cant_happen(), (NODE*)NULL) : \
 			(_t->type == Node_param_list ? \
 			  r_tree_eval(_t, iscond) : \
 			((_t->type == Node_val && (_t->flags&INTLSTR)) ? \
 				r_force_string(_t) : \
 			(_t->type == Node_val ? _t : \
 			(_t->type == Node_var && \
-			 (_t->flags & UNINITIALIZED) == 0 ? _t->var_value : \
+			 ! var_uninitialized(_t) ? _t->var_value : \
 			 r_tree_eval(_t, iscond))))))
 #endif /* __GNUC__ */
 #endif /* not MEMDEBUG */
@@ -805,20 +908,29 @@ extern const char casetable[];	/* for case-independent regexp matching */
 #define	cant_happen()	r_fatal("internal error line %d, file: %s", \
 				__LINE__, __FILE__)
 
+/*
+ * For SunOS 4.1.x which is pre-Standard C, `realloc' doesn't
+ * accept NULL. Sigh.  The check must be done for both cases,
+ * since could be using GCC but with stock C library. Sigh, again.
+ */
 #ifdef HAVE_STRINGIZE
 #define	emalloc(var,ty,x,str)	(void)((var=(ty)malloc((MALLOC_ARG_T)(x))) ||\
 				 (fatal(_("%s: %s: can't allocate %ld bytes of memory (%s)"),\
 					(str), #var, (long) (x), strerror(errno)),0))
-#define	erealloc(var,ty,x,str)	(void)((var=(ty)realloc((char *)var,\
-						  (MALLOC_ARG_T)(x))) ||\
+#define	erealloc(var,ty,x,str)	(void)((var = ((var == NULL) \
+					? (ty)malloc((MALLOC_ARG_T)(x)) \
+					: (ty)realloc((char *)var, (MALLOC_ARG_T)(x))) ) \
+				||\
 				 (fatal(_("%s: %s: can't allocate %ld bytes of memory (%s)"),\
 					(str), #var, (long) (x), strerror(errno)),0))
 #else /* HAVE_STRINGIZE */
 #define	emalloc(var,ty,x,str)	(void)((var=(ty)malloc((MALLOC_ARG_T)(x))) ||\
 				 (fatal(_("%s: %s: can't allocate %ld bytes of memory (%s)"),\
 					(str), "var", (long) (x), strerror(errno)),0))
-#define	erealloc(var,ty,x,str)	(void)((var=(ty)realloc((char *)var,\
-						  (MALLOC_ARG_T)(x))) ||\
+#define	erealloc(var,ty,x,str)	(void)((var = ((var == NULL) \
+					? (ty)malloc((MALLOC_ARG_T)(x)) \
+					: (ty)realloc((char *)var, (MALLOC_ARG_T)(x))) ) \
+				||\
 				 (fatal(_("%s: %s: can't allocate %ld bytes of memory (%s)"),\
 					(str), "var", (long) (x), strerror(errno)),0))
 #endif /* HAVE_STRINGIZE */
@@ -867,6 +979,8 @@ extern double _msc51bug;
 /* ------------- Function prototypes or defs (as appropriate) ------------- */
 
 /* array.c */
+extern NODE *get_actual P((NODE *symbol, int canfatal));
+extern char *array_vname P((const NODE *symbol));
 extern void array_init P((void));
 extern NODE *concat_exp P((NODE *tree));
 extern void assoc_clear P((NODE *symbol));
@@ -874,11 +988,12 @@ extern NODE *in_array P((NODE *symbol, NODE *subs));
 extern NODE **assoc_lookup P((NODE *symbol, NODE *subs, int reference));
 extern void do_delete P((NODE *symbol, NODE *tree));
 extern void do_delete_loop P((NODE *symbol, NODE *tree));
+extern void set_SUBSEP P((void));
 extern NODE *assoc_dump P((NODE *symbol));
 extern NODE *do_adump P((NODE *tree));
 extern NODE *do_asort P((NODE *tree));
 extern NODE *do_asorti P((NODE *tree));
-extern unsigned long (*hash)P((const char *s, size_t len, unsigned long hsize));
+extern unsigned long (*hash)P((const char *s, size_t len, unsigned long hsize, size_t *code));
 /* awkgram.c */
 extern char *tokexpand P((void));
 extern NODE *node P((NODE *left, NODETYPE op, NODE *right));
@@ -892,6 +1007,9 @@ extern void release_all_vars P((void));
 extern const char *getfname P((NODE *(*)(NODE *)));
 extern NODE *stopme P((NODE *tree));
 extern void shadow_funcs P((void));
+extern int check_special P((const char *name));
+extern void register_deferred_variable P((const char *name,
+					  NODE *(*load_func)(void)));
 /* builtin.c */
 extern double double_to_int P((double d));
 extern NODE *do_exp P((NODE *tree));
@@ -952,11 +1070,15 @@ extern void set_BINMODE P((void));
 extern void set_LINT P((void));
 extern void set_TEXTDOMAIN P((void));
 extern void update_ERRNO P((void));
+extern void update_ERRNO_saved P((int));
 extern const char *redflags2str P((int));
 extern const char *flags2str P((int));
 extern const char *genflags2str P((int flagval, const struct flagtab *tab));
 extern const char *nodetype2str P((NODETYPE type));
 extern NODE *assign_val P((NODE **lhs_p, NODE *rhs));
+extern void load_casetable P((void));
+extern size_t get_curfunc_arg_count P((void));
+extern AWKNUM calc_exp P((AWKNUM x1, AWKNUM x2));
 #ifdef PROFILING
 extern void dump_fcall_stack P((FILE *fp));
 #endif
@@ -965,6 +1087,9 @@ NODE *do_ext P((NODE *));
 #ifdef DYNAMIC
 void make_builtin P((char *, NODE *(*)(NODE *), int));
 NODE *get_argument P((NODE *, int));
+NODE *get_actual_argument P((NODE *, unsigned int, int, int));
+#define get_scalar_argument(t, i, opt)  get_actual_argument((t), (i), (opt), FALSE)
+#define get_array_argument(t, i, opt)   get_actual_argument((t), (i), (opt), TRUE)
 void set_value P((NODE *));
 #endif
 /* field.c */
@@ -991,14 +1116,25 @@ extern void os_restore_mode P((int fd));
 extern size_t optimal_bufsize P((int fd, struct stat *sbuf));
 extern int ispath P((const char *file));
 extern int isdirpunct P((int c));
+
+#if HAVE_MEMCPY_ULONG
+extern char *memcpy_ulong P((char *dest, const char *src, unsigned long l));
+#define memcpy memcpy_ulong
+#endif
+#if HAVE_MEMSET_ULONG
+extern void *memset_ulong P((void *dest, int val, unsigned long l));
+#define memset memset_ulong
+#endif
+
 /* io.c */
+extern void register_open_hook P((void *(*open_func)(IOBUF *)));
 extern void set_FNR P((void));
 extern void set_NR P((void));
 extern void do_input P((void));
 extern struct redirect *redirect P((NODE *tree, int *errflg));
 extern NODE *do_close P((NODE *tree));
 extern int flush_io P((void));
-extern int close_io P((void));
+extern int close_io P((int *stdio_problem));
 extern int devopen P((const char *name, const char *mode));
 extern int pathopen P((const char *file));
 extern NODE *do_getline P((NODE *tree));
@@ -1006,9 +1142,8 @@ extern void do_nextfile P((void)) ATTRIBUTE_NORETURN;
 extern struct redirect *getredirect P((const char *str, int len));
 /* main.c */
 extern int main P((int argc, char **argv));
-extern NODE *load_environ P((void));
-extern NODE *load_procinfo P((void));
 extern int arg_assign P((char *arg, int initing));
+extern int is_std_var P((const char *var));
 /* msg.c */
 extern void err P((const char *s, const char *emsg, va_list argp)) ATTRIBUTE_PRINTF(2, 0);
 #if _MSC_VER == 510
@@ -1025,6 +1160,7 @@ extern void error (const char *mesg, ...) ATTRIBUTE_PRINTF_1;
 extern void warning (const char *mesg, ...) ATTRIBUTE_PRINTF_1;
 extern void set_loc (const char *file, int line);
 extern void r_fatal (const char *mesg, ...) ATTRIBUTE_PRINTF_1 ATTRIBUTE_NORETURN;
+ATTRIBUTE_EXPORTED 
 #if __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 2)
 extern void (*lintfunc) (const char *mesg, ...) ATTRIBUTE_PRINTF_1;
 #else
@@ -1062,17 +1198,42 @@ extern void freenode P((NODE *it));
 #endif
 extern void unref P((NODE *tmp));
 extern int parse_escape P((const char **string_ptr));
+#ifdef MBS_SUPPORT
+extern NODE *str2wstr P((NODE *n, size_t **ptr));
+#define force_wstring(n)	str2wstr(n, NULL)
+extern const wchar_t *wstrstr P((const wchar_t *haystack, size_t hs_len, const wchar_t *needle, size_t needle_len));
+extern const wchar_t *wcasestrstr P((const wchar_t *haystack, size_t hs_len, const wchar_t *needle, size_t needle_len));
+extern void free_wstr P((NODE *n));
+#else
+#define free_wstr(NODE)	/* empty */
+#endif
 /* re.c */
-extern Regexp *make_regexp P((const char *s, size_t len, int ignorecase));
-extern int research P((Regexp *rp, const char *str, int start,
-		       size_t len, int need_start));
+extern Regexp *make_regexp P((const char *s, size_t len, int ignorecase, int dfa));
+extern int research P((Regexp *rp, char *str, int start,
+		       size_t len, int flags));
 extern void refree P((Regexp *rp));
 extern void reg_error P((const char *s));
 extern Regexp *re_update P((NODE *t));
 extern void resyntax P((int syntax));
 extern void resetup P((void));
+extern int avoid_dfa P((NODE *re, char *str, size_t len));	/* temporary */
 extern int reisstring P((const char *text, size_t len, Regexp *re, const char *buf));
 extern int remaybelong P((const char *text, size_t len));
+extern int isnondecimal P((const char *str, int use_locale));
+
+/* floatcomp.c */
+#ifdef VMS	/* VMS linker weirdness? */
+#define Ceil	gawk_ceil
+#define Floor	gawk_floor
+#endif
+
+extern AWKNUM Floor P((AWKNUM n));
+extern AWKNUM Ceil P((AWKNUM n));
+#ifdef HAVE_UINTMAX_T
+extern uintmax_t adjust_uint P((uintmax_t n));
+#else
+#define adjust_uint(n) (n)
+#endif
 
 /* strncasecmp.c */
 #ifndef BROKEN_STRNCASECMP
@@ -1098,9 +1259,36 @@ extern char *tempnam P((const char *path, const char *base));
 #include <sys/wait.h>
 #endif
 #ifndef WEXITSTATUS
+#if defined(_MSC_VER) || defined(VMS)
+#define WEXITSTATUS(stat_val) (stat_val)
+#else /* ! (defined(_MSC_VER) || defined(VMS)) */
 #define WEXITSTATUS(stat_val) ((((unsigned) (stat_val)) >> 8) & 0xFF)
-#endif
+#endif /* ! (defined(_MSC_VER) || defined(VMS)) */
+#endif /* WEXITSTATUS */
 
 #ifndef STATIC
 #define STATIC static
 #endif
+
+/* EXIT_SUCCESS and EXIT_FAILURE normally come from <stdlib.h> */
+#ifndef EXIT_SUCCESS
+# define EXIT_SUCCESS 0
+#endif
+#ifndef EXIT_FAILURE
+# define EXIT_FAILURE 1
+#endif
+/* EXIT_FATAL is specific to gawk, not part of Standard C */
+#ifndef EXIT_FATAL
+# define EXIT_FATAL   2
+#endif
+
+/* For z/OS, from Dave Pitts. EXIT_FAILURE is normally 8, make it 1. */
+#ifdef ZOS_USS
+#undef DYNAMIC
+
+#ifdef EXIT_FAILURE
+#undef EXIT_FAILURE
+#endif
+
+#define EXIT_FAILURE 1
+#endif
