@@ -11,14 +11,14 @@
  */
 
 /* 
- * Copyright (C) 1995-2002 the Free Software Foundation, Inc.
+ * Copyright (C) 1995-2004, 2008, 2009 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  * 
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  * 
  * GAWK is distributed in the hope that it will be useful,
@@ -28,7 +28,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 /* for MIPS RiscOS, from Nelson H. F. Beebe, beebe@math.utah.edu */
@@ -95,3 +95,19 @@
 #ifdef USE_INCLUDED_STRFTIME
 #undef HAVE_STRFTIME
 #endif
+
+/* For HP/UX with gcc */
+#if defined(hpux) || defined(_HPUX_SOURCE)
+#undef HAVE_TZSET
+#define HAVE_TZSET 1
+#define _TZSET 1
+#endif
+
+/* For z/OS, from Dave Pitts */
+#ifdef ZOS_USS
+#undef HAVE_DLFCN_H
+#endif
+
+#if !defined(NO_LIBSIGSEGV) && !defined(DJGPP) && !defined(VMS)
+#define HAVE_SIGSEGV_H 1
+#endif
