@@ -69,8 +69,8 @@ register const char **ptr;
 	}
 
 	if (
-#if ENABLE_NLS && defined(HAVE_LOCALE_H)
-	loc.decimal_point != NULL
+#if defined(HAVE_LOCALE_H)
+	loc.decimal_point != NULL && do_posix
 			? *s == loc.decimal_point[0]
 			: *s == '.'
 #else
